//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  mem.cpp
//! \brief Implement online memory

#include "defs.hpp"
#include "../trivia/vector_ops.hpp"
#include <cmath>
#include <algorithm>

#include "olf.hpp"
#include "ptr.hpp"
#include "obj.hpp"
#include "mem.hpp"
#include "fsel.hpp"
#include "loop.hpp"


//**************************************************************************************************
//*  Singleton instances
//**************************************************************************************************

namespace NOlm {

    vector<POlo> MOM;
    vector<POlo> BAK;
    bool overwrite;


//**************************************************************************************************
//*  Read access to online memory
//**************************************************************************************************

    //! Returns number of files in online memory.
    const int mem_size()
    {
        return MOM.size();
    }

    //! Return pointer to file with index k.
    const POlo mem_get( int k )
    {
        if ( k<0 )
            throw "negative number " + S(k) + " cannot be a file index";
        if ( k>=MOM.size() )
            throw "file index " + S(k) + " not in online memory";
        return MOM[k];
    }

    //! Return pointer to file with index k, provided it contains data.
    const POld mem_get_D( int k )
    {
        POld ret = P2D( mem_get( k ) );
        if( !ret )
            throw "file " + S(k) + " is not a data file";
        return ret;
    }

    //! Return pointer to file with index k, provided it contains curves.
    const POlc mem_get_C( int k )
    {
        POlc ret = P2C( mem_get( k ) );
        if( !ret )
            throw "file " + S(k) + " is not a curve file";
        return ret;
    }


//**************************************************************************************************
//*  Modify online memory, and adapt file NSel::selection
//**************************************************************************************************

    //! Store file f in online memory. If overwrite and k!=-1, overwrite file k.
    void mem_store( POlo f, int k )
    {
        if( overwrite && k!=-1 ){
            // replace file k
            if( k>=MOM.size() )
                throw S("BUG: cannot overwrite inexistent file");
            MOM[k] = f;
        } else {
            // append at end of online memory
            MOM.push_back( f );
            NSel::sel_push_back( MOM.size()-1 );
        }
    }

    //! Remove file k from online memory.
    void mem_del( int k )
    {
        MOM.erase(MOM.begin()+k);
        /* TODO update old NSel::selection(s) after file deletion
        for( int i=0; i<FNSel::SelOld.size(); ++i )
            if ( FNSel::SelOld[i] > k )
                --FNSel::SelOld[i];
        */
    }

    //! Append copy of NSel::selected files to online memory.
    void mem_copy()
    {
        FileIterator fiter(NSel::sel());
        while ( ROlo fin=fiter.get() ){
            POlo fout( fin->new_olo( false ) );
            fout->copy_mainvec( fin );
            mem_store( fout );
        }
    }

    //! Move files to end of online memory.
    void mem_xchange()
    {
        mem_copy();
        NSel::sel_delete();
    }

//**************************************************************************************************
//*  Backup and retrieve
//**************************************************************************************************

    //! Copy NSel::selected files to backup.

    void mem_backup_store()
    {
        BAK.clear();
        FileIterator fiter(NSel::sel());
        while ( ROlo fin=fiter.get() ){
            POlo fout( fin->new_olo( false ) );
            fout->copy_mainvec( fin );
            BAK.push_back( fout );
        }
    }

    //! Retrieve files from backup.

    void mem_backup_retrieve()
    {
        if( !BAK.size() )
            throw S("no files in backup");
        NSel::sel_collect_begin();
        for( auto b: BAK )
            mem_store( b, -1 );
        NSel::sel_collect_end();
        BAK.clear();
    }

} // NOlm
