//**************************************************************************************************
//* libread-plus: the command-line dialog utility.
//* macro:
//* (C) Joachim Wuttke 2001, 2012-
//* http://apps.jcns.fz-juelich.de
//**************************************************************************************************

#include <cstdio>
#include <cstring>
#include <wordexp.h>
#include <boost/algorithm/string/trim.hpp>

#include <string>
#include <iostream>
#include <fstream>
#include <map>
#include <deque>

#include "../trivia/string_ops.hpp"
#include "../trivia/file_ops.hpp"
#include "readln.hpp"
#include "macro.hpp"

using std::string;
using std::ifstream;
using std::cout;

namespace NMacro {
    // internals, used only in this file:
    std::deque<string> stack;

    // the following are only called indirectly through macros:
    void metacmd_incl( string r );

    // auxiliary:
    string wordexp_cpp_unique( const string& s );
    string string_extract_quoted( string *in );
    bool is_separator( char c );
}


//**************************************************************************************************
//*  API
//**************************************************************************************************


//! Take a line from the stack, for use in readln or readwd.

string NMacro::readln( const string& prompt, ifstream* F, int* lineno )
{
    // Obtain input line.
    string in;
    if ( stack.empty() ) {
        if( F ) {
            ++(*lineno);
            if( !(getline(*F, in)) )
                return "EOF";
        } else {
            in = NReadln::readline( prompt );
        }
    } else {
        in = *(stack.begin());
        stack.pop_front();
    }
    boost::trim( in );
    if      ( in=="" ) {
        return "";
    } else if ( in[0]=='#' ) {
        return "";
    } else if ( in[0]=='\\' ) {  // macro \..
        return exemac( in );
    } else if ( in=="''" ) { // needed ?
        return "\\EMPTYSTRING";  // explicit empty string, not default
    }
    if ( is_separator( in[0] ) )
        throw "BUG: invalid leading space in input [" + in + "]";
    return in;
}


//! Take word from stack. Prompt for input if stack is empty.

string NMacro::readwd( const string& prompt )
{
    string in = readln( prompt );
    string::size_type j = in.find_first_of( " \t" );
    if ( j!=string::npos ) {
        push_remainder( in.substr(j) );
        return in.substr( 0, j );
    }
    return in;
}


//! Clear input stack.

void NMacro::clear()
{
    stack.clear();
}


//! Execute metacommand or substitute macro.

string NMacro::exemac( const string& in )
{
    string cmd, arg;
    triv::string_extract_word( in, &cmd, &arg );
    std::map<string,string>::iterator pos2;
    if        ( cmd=="\\?" || cmd=="\\h" ) {
        cout <<
            "  meta commands:\n"
            "    \\i <file>               include file\n"
            "    \\q                      escape to main menu\n"
            "    \\wh                     write history\n";
        return "";
    } else if ( cmd=="\\" ) {
        push_remainder( arg );
        return "\\DEFAULT";
    } else if ( in=="\\q" ) {
        throw string( "user escape to main menu" );
    } else if ( cmd=="\\wh" ) {
        NReadln::write_history( "history" );
        push_remainder( arg );
        return "";
    } else if ( cmd=="\\i" ) {
        metacmd_incl( arg );
        return "";
    }
    throw "macro " + cmd + " undefined";
}


//**************************************************************************************************
//  internal fuctions
//**************************************************************************************************


//! Push _rem_ back on the input stack.

void NMacro::push_remainder( const string& rem )
{
    if ( rem=="" )
        return; // nothing to do
    if ( !is_separator( rem[0] ) )
        throw "missing separator at beginning of [" + rem + "]";
    string::size_type j = rem.find_first_not_of( " \t" );
    if ( j!=string::npos )
        stack.push_front( rem.substr(j) );
}


//! Returns one file name, obtained by bash expansion of argument.

string NMacro::wordexp_cpp_unique( const string& s )
{
    wordexp_t p;
    if( wordexp(s.c_str(), &p, 0) )
        throw "cannot expand " + s;
    if( p.we_wordc!=1 )
        throw "expansion of " + s + " not unique";
    string ret = p.we_wordv[0];
    wordfree( &p );
    return ret;
}


//! Include a script.

//! Input:  command stack r, with script filename on top.
//! Output: command stack, with script filename substituted by script contents.

void NMacro::metacmd_incl( string r )
{
    string ftxt;
    triv::string_extract_word( r, &ftxt, &r );
    if( r!="" )
        stack.push_front( r );
    string fnam = NMacro::wordexp_cpp_unique( ftxt );
    FILE *fd;
    if( !(fd = fopen( fnam.c_str(), "r" )) )
        throw "cannot open file " + fnam;
    string s;
    while ( triv::freadln( fd, &s ) )
        stack.push_back( s );
    fclose( fd );
}


//**************************************************************************************************
//  Auxiliary string operations
//**************************************************************************************************

//! Test whether argument is a blank or a tab.

bool NMacro::is_separator( char c )
{
    return string(" \t").find_first_of( c )!=string::npos;
}
